<!--

    vcflib page generated with

      ./scripts/bin2md.rb scripts/bin2mdidx-template.erb

    >>> from pytest.rtest import run_stdout, head, cat

-->

## <%= cmd %>

<%= descr %>

```sh

>>> cat("<%= help_cmd %>")
<%= pydoc_full %>

```

### Source code

[<%= cmd %>.cpp](https://github.com/vcflib/vcflib/blob/master/src/<%= cmd %>.cpp)

---
